"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.eventToObject = exports.loadScript = exports.toModuleUrl = exports.expose = exports.setUpMuiFixModule = void 0;
const sucrase_1 = require("sucrase");
const muiStyleFix = `
// This is a specific 'fix' for the notebook only, since its fontsize is non-16
import React from 'react';
// import PropTypes from 'prop-types';
// import { MuiThemeProvider, createMuiTheme } from '@material-ui/core/styles';
import Typography from '@mui/material/Typography';
import { createTheme, ThemeProvider } from '@mui/material/styles';

const theme = createTheme({
  typography: {
    // Tell Material-UI what the font-size on the html element is.
    htmlFontSize: 10,
    useNextVariants: true,
},
});

function FontSizeTheme({ children }) {
    return (
        <ThemeProvider theme={theme}>
            <Typography component="span">{children}</Typography>
        </ThemeProvider>
    );
}

// FontSizeTheme.propTypes = {
//     children: PropTypes.node.isRequired,
// };

export
function styleWrapper(element) {
    if(window.Jupyter) {
        return <FontSizeTheme>{element}</FontSizeTheme>;
    } else {
        return element;
    }
}
`;
function setUpMuiFixModule() {
    return __awaiter(this, void 0, void 0, function* () {
        const code = sucrase_1.transform(muiStyleFix, {
            transforms: ["jsx", "typescript"],
            filePath: "muifix.tsx",
        }).code;
        let url = URL.createObjectURL(new Blob([code], { type: "text/javascript" }));
        // @ts-ignore
        return yield importShim(url);
    });
}
exports.setUpMuiFixModule = setUpMuiFixModule;
function expose(module) {
    const id = "_ipyreact_" + Math.random().toString(36);
    // @ts-ignore
    window[id] = module;
    const names = Object.keys(module)
        .filter((n) => n !== "default")
        .join(", ");
    return toModuleUrl(`
        const { ${names} } = window["${id}"];
        export default window["${id}"].default;
        delete window["${id}"];
        export { ${names} };`);
}
exports.expose = expose;
function toModuleUrl(code) {
    return URL.createObjectURL(new Blob([code], { type: "text/javascript" }));
}
exports.toModuleUrl = toModuleUrl;
function loadScript(type, src) {
    return __awaiter(this, void 0, void 0, function* () {
        const script = document.createElement("script");
        script.type = type;
        script.src = src;
        script.defer = true;
        document.head.appendChild(script);
        return new Promise((resolve, reject) => {
            script.onload = () => {
                resolve();
            };
            script.onerror = () => {
                reject();
            };
        });
    });
}
exports.loadScript = loadScript;
// based on https://stackoverflow.com/a/58416333/5397207
function pickSerializable(object, depth = 0, max_depth = 2) {
    // change max_depth to see more levels, for a touch event, 2 is good
    if (depth > max_depth)
        return "Object";
    const obj = {};
    for (let key in object) {
        let value = object[key];
        if (value instanceof Window)
            value = "Window";
        else if (value && value.getModifierState)
            value = pickSerializable(value, depth + 1, max_depth);
        else {
            // test if serializable
            try {
                JSON.stringify(value);
            }
            catch (e) {
                value = "Object";
            }
        }
        obj[key] = value;
    }
    return obj;
}
function eventToObject(event) {
    if (event instanceof Event || (event && event.getModifierState)) {
        return pickSerializable(event);
    }
    return event;
}
exports.eventToObject = eventToObject;
//# sourceMappingURL=utils.js.map